@@ -6,19 +6,22 @@ import (
 	"fmt"
 	"io/ioutil"
 	"os"
+	"reflect"
 	"testing"
 
 	imageregistryv1 "github.com/openshift/api/imageregistry/v1"
 	networkv1 "github.com/openshift/api/network/v1"
-	v1 "k8s.io/api/core/v1"
+	corev1 "k8s.io/api/core/v1"
+	"k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
+	apixv1beta1clientfake "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset/fake"
 	policyv1beta1 "k8s.io/api/policy/v1beta1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/apimachinery/pkg/runtime/serializer/yaml"
-	"k8s.io/apimachinery/pkg/util/intstr"
 	dynamicfake "k8s.io/client-go/dynamic/fake"
+	"k8s.io/apimachinery/pkg/util/intstr"
 	kubefake "k8s.io/client-go/kubernetes/fake"
 	"k8s.io/klog"
 
@@ -67,7 +70,7 @@ func TestConfigMapAnonymizer(t *testing.T) {
 			defer f.Close()
 			bts, err := ioutil.ReadAll(f)
 			mustNotFail(t, err, "error reading test data file. %+v")
-			var cml *v1.ConfigMapList
+			var cml *corev1.ConfigMapList
 			mustNotFail(t, json.Unmarshal([]byte(bts), &cml), "error unmarshalling json %+v")
 			cm := findMap(cml, tt.configMapName)
 			mustNotFail(t, cm != nil, "haven't found a ConfigMap %+v")
@@ -172,21 +175,20 @@ func TestGatherClusterPruner(t *testing.T) {
 			test.evalOutput(t, obj.(*imageregistryv1.ImagePruner))
 		})
 	}
-
 }
 
-func TestGatherPodDisruptionBudgets(t *testing.T) {
+func TestGatherPodDisruptionBudgets(t *testing.T){
 	coreClient := kubefake.NewSimpleClientset()
 
 	fakeNamespace := "fake-namespace"
 
 	// name -> MinAvailabel
-	fakePDBs := map[string]string{
-		"pdb-four":  "4",
+	fakePDBs := map[string]string {
+		"pdb-four": "4",
 		"pdb-eight": "8",
-		"pdb-ten":   "10",
+		"pdb-ten": "10",
 	}
-	for name, minAvailable := range fakePDBs {
+	for name, minAvailable := range fakePDBs{
 		_, err := coreClient.PolicyV1beta1().
 			PodDisruptionBudgets(fakeNamespace).
 			Create(context.Background(), &policyv1beta1.PodDisruptionBudget{
@@ -220,7 +222,7 @@ func TestGatherPodDisruptionBudgets(t *testing.T) {
 		}
 		name := pdba.PodDisruptionBudget.ObjectMeta.Name
 		minAvailable := pdba.PodDisruptionBudget.Spec.MinAvailable.StrVal
-		if pdba.PodDisruptionBudget.Spec.MinAvailable.StrVal != fakePDBs[name] {
+		if pdba.PodDisruptionBudget.Spec.MinAvailable.StrVal !=  fakePDBs[name] {
 			t.Fatalf("pdb item has mismatched MinAvailable value, %q != %q", fakePDBs[name], minAvailable)
 		}
 	}
@@ -364,41 +366,173 @@ func TestGatherClusterImageRegistry(t *testing.T) {
 			test.evalOutput(t, obj.(*imageregistryv1.Config))
 		})
 	}
-
 }
-func TestGatherMachineSet(t *testing.T) {
-	var machineSetYAML = `
-apiVersion: machine.openshift.io/v1beta1
-kind: MachineSet
-metadata:
-    name: test-worker
-`
-	gvr := schema.GroupVersionResource{Group: "machine.openshift.io", Version: "v1beta1", Resource: "machinesets"}
-	client := dynamicfake.NewSimpleDynamicClient(runtime.NewScheme())
-	decUnstructured := yaml.NewDecodingSerializer(unstructured.UnstructuredJSONScheme)
 
-	testMachineSet := &unstructured.Unstructured{}
+func TestGatherContainerImages(t *testing.T) {
+	const fakeNamespace = "fake-namespace"
+	const fakeOpenshiftNamespace = "openshift-fake-namespace"
 
-	_, _, err := decUnstructured.Decode([]byte(machineSetYAML), nil, testMachineSet)
-	if err != nil {
-		t.Fatal("unable to decode machineset ", err)
-	}
-	_, err = client.Resource(gvr).Create(context.Background(), testMachineSet, metav1.CreateOptions{})
-	if err != nil {
-		t.Fatal("unable to create fake machineset ", err)
+	mockContainers := []string{
+		"registry.redhat.io/1",
+		"registry.redhat.io/2",
+		"registry.redhat.io/3",
 	}
 
-	gatherer := &Gatherer{dynamicClient: client}
-	records, errs := GatherMachineSet(gatherer)()
+	expected := ContainerInfo{
+		Images: ContainerImageSet{
+			0: "registry.redhat.io/1",
+			1: "registry.redhat.io/2",
+			2: "registry.redhat.io/3",
+		},
+		Containers: PodsWithAge{
+			"0001-01": RunningImages{
+				0: 1,
+				1: 1,
+				2: 1,
+			},
+		},
+	}
+
+	coreClient := kubefake.NewSimpleClientset()
+	for index, containerImage := range mockContainers {
+		_, err := coreClient.CoreV1().
+			Pods(fakeNamespace).
+			Create(context.Background(), &corev1.Pod{
+				ObjectMeta: metav1.ObjectMeta{
+					Namespace: fakeNamespace,
+					Name:      fmt.Sprintf("pod%d", index),
+				},
+				Spec: corev1.PodSpec{
+					Containers: []corev1.Container{
+						{
+							Name:  fmt.Sprintf("container%d", index),
+							Image: containerImage,
+						},
+					},
+				},
+				Status: corev1.PodStatus{
+					Phase: corev1.PodRunning,
+				},
+			}, metav1.CreateOptions{})
+		if err != nil {
+			t.Fatal("unable to create fake pod")
+		}
+	}
+
+	const numberOfCrashlooping = 10
+	expectedRecords := make([]string, numberOfCrashlooping)
+	for i := 0; i < numberOfCrashlooping; i++ {
+		podName := fmt.Sprintf("crashlooping%d", i)
+		_, err := coreClient.CoreV1().
+			Pods(fakeOpenshiftNamespace).
+			Create(context.Background(), &corev1.Pod{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: podName,
+				},
+				Status: corev1.PodStatus{
+					ContainerStatuses: []corev1.ContainerStatus{
+						{
+							RestartCount: int32(numberOfCrashlooping - i),
+							LastTerminationState: corev1.ContainerState{
+								Terminated: &corev1.ContainerStateTerminated{
+									ExitCode: int32(i + 1),
+								},
+							},
+						},
+					},
+				},
+			}, metav1.CreateOptions{})
+		if err != nil {
+			t.Fatal("unable to create fake pod")
+		}
+		expectedRecords[i] = fmt.Sprintf("config/pod/%s/%s", fakeOpenshiftNamespace, podName)
+	}
+
+	gatherer := &Gatherer{coreClient: coreClient.CoreV1()}
+
+	records, errs := GatherContainerImages(gatherer)()
 	if len(errs) > 0 {
 		t.Errorf("unexpected errors: %#v", errs)
 		return
 	}
-	if len(records) != 1 {
-		t.Fatalf("unexpected number or records %d", len(records))
+
+	var containerInfo *ContainerInfo = nil
+	for _, rec := range records {
+		if rec.Name == "config/running_containers" {
+			anonymizer, ok := rec.Item.(record.JSONMarshaller)
+			if !ok {
+				t.Fatal("reported running containers item has invalid type")
+			}
+
+			containers, ok := anonymizer.Object.(ContainerInfo)
+			if !ok {
+				t.Fatal("anonymized running containers data have wrong type")
+			}
+
+			containerInfo = &containers
+		}
 	}
-	if records[0].Name != "machinesets/test-worker" {
-		t.Fatalf("unexpected machineset name %s", records[0].Name)
+
+	if containerInfo == nil {
+		t.Fatal("container info has not been reported")
+	}
+
+	if !reflect.DeepEqual(*containerInfo, expected) {
+		t.Fatalf("unexpected result: %#v", *containerInfo)
+	}
+
+	for _, expectedRecordName := range expectedRecords {
+		wasReported := false
+		for _, reportedRecord := range records {
+			if reportedRecord.Name == expectedRecordName {
+				wasReported = true
+				break
+			}
+		}
+		if !wasReported {
+			t.Fatalf("expected record '%s' was not reported", expectedRecordName)
+		}
+	}
+}
+
+func TestCollectVolumeSnapshotCRD(t *testing.T) {
+	expectedRecords := map[string]v1beta1.CustomResourceDefinition{
+		"config/crd/volumesnapshots.snapshot.storage.k8s.io":        {ObjectMeta: metav1.ObjectMeta{Name: "volumesnapshots.snapshot.storage.k8s.io"}},
+		"config/crd/volumesnapshotcontents.snapshot.storage.k8s.io": {ObjectMeta: metav1.ObjectMeta{Name: "volumesnapshotcontents.snapshot.storage.k8s.io"}},
+	}
+
+	crdNames := []string{
+		"unrelated.custom.resource.definition.k8s.io",
+		"volumesnapshots.snapshot.storage.k8s.io",
+		"volumesnapshotcontents.snapshot.storage.k8s.io",
+		"another.irrelevant.custom.resource.definition.k8s.io",
+		"this.should.not.be.gathered.k8s.io",
+	}
+
+	crdClientset := apixv1beta1clientfake.NewSimpleClientset()
+
+	for _, name := range crdNames {
+		crdClientset.ApiextensionsV1beta1().CustomResourceDefinitions().Create(context.Background(), &v1beta1.CustomResourceDefinition{
+			ObjectMeta: metav1.ObjectMeta{Name: name},
+		}, metav1.CreateOptions{})
+	}
+
+	gatherer := &Gatherer{crdClient: crdClientset.ApiextensionsV1beta1()}
+	records, errs := GatherCRD(gatherer)()
+	if len(errs) != 0 {
+		t.Fatalf("gather CRDs resulted in error: %#v", errs)
+	}
+
+	if len(records) != len(expectedRecords) {
+		t.Fatalf("unexpected number of records gathered: %d (expected %d)", len(records), len(expectedRecords))
+	}
+
+	for _, rec := range records {
+		if expectedItem, ok := expectedRecords[rec.Name]; !ok {
+			t.Fatalf("unexpected gathered record name: %q", rec.Name)
+		} else if reflect.DeepEqual(rec.Item, expectedItem) {
+			t.Fatalf("gathered record %q has different item value than unexpected", rec.Name)
+		}
 	}
 }
 
@@ -459,6 +593,42 @@ func TestGatherHostSubnet(t *testing.T) {
 	}
 }
 
+func TestGatherMachineSet(t *testing.T) {
+	var machineSetYAML = `
+apiVersion: machine.openshift.io/v1beta1
+kind: MachineSet
+metadata:
+    name: test-worker
+`
+	gvr := schema.GroupVersionResource{Group: "machine.openshift.io", Version: "v1beta1", Resource: "machinesets"}
+	client := dynamicfake.NewSimpleDynamicClient(runtime.NewScheme())
+	decUnstructured := yaml.NewDecodingSerializer(unstructured.UnstructuredJSONScheme)
+
+	testMachineSet := &unstructured.Unstructured{}
+
+	_, _, err := decUnstructured.Decode([]byte(machineSetYAML), nil, testMachineSet)
+	if err != nil {
+		t.Fatal("unable to decode machineset ", err)
+	}
+	_, err = client.Resource(gvr).Create(context.Background(), testMachineSet, metav1.CreateOptions{})
+	if err != nil {
+		t.Fatal("unable to create fake machineset ", err)
+	}
+
+	gatherer := &Gatherer{dynamicClient: client}
+	records, errs := GatherMachineSet(gatherer)()
+	if len(errs) > 0 {
+		t.Errorf("unexpected errors: %#v", errs)
+		return
+	}
+	if len(records) != 1 {
+		t.Fatalf("unexpected number or records %d", len(records))
+	}
+	if records[0].Name != "machinesets/test-worker" {
+		t.Fatalf("unexpected machineset name %s", records[0].Name)
+	}
+}
+
 func ExampleGatherMostRecentMetrics_Test() {
 	b, err := ExampleMostRecentMetrics()
 	if err != nil {
@@ -498,7 +668,7 @@ func mustNotFail(t *testing.T, err interface{}, fmtstr string) {
 	}
 }
 
-func findMap(cml *v1.ConfigMapList, name string) *v1.ConfigMap {
+func findMap(cml *corev1.ConfigMapList, name string) *corev1.ConfigMap {
 	for _, it := range cml.Items {
 		if it.Name == name {
 			return &it
